@@ -1,6 +1,6 @@
 #  sfzen/sfz_elems.py
 #
-#  Copyright 2024 liyang <liyang@veronica>
+#  Copyright 2024 Leon Dionne <ldionne@dridesign.sh.cn>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -22,9 +22,10 @@ Classes which are instantiated when parsing an .sfz file.
 All of these classes are constructed from a lark parser tree Token.
 """
 import re, logging
-from os import symlink, link, sep as path_separator
-from os.path import abspath, exists, join, relpath
+from os import unlink, symlink, link as hardlink, sep as path_separator
+from os.path import abspath, exists, islink, join, relpath, realpath
 from shutil import move, copy2 as copy
+from copy import deepcopy
 from functools import cached_property, reduce
 try:
 	from functools import cache
@@ -90,6 +91,24 @@ class Header(_SFZElement):
 		self._subheaders = []
 		self._opcodes = {}
 
+	def __deepcopy__(self, memo):
+		cls = self.__class__
+		result = cls.__new__(cls)
+		result._subheaders = [
+			deepcopy(subheader, memo) \
+			for subheader in self._subheaders
+		]
+		result._opcodes = {
+			key:deepcopy(opcode, memo) \
+			for key, opcode in self._opcodes.items()
+		}
+		for subheader in result._subheaders:
+			subheader.parent = result
+		for opcode in result._opcodes.values():
+			opcode.parent = result
+		memo[id(self)] = result
+		return result
+
 	def may_contain(self, _):
 		"""
 		This function is used to determine if a header being parsed is a child of the
@@ -100,16 +119,22 @@ class Header(_SFZElement):
 	def append_opcode(self, opcode):
 		"""
 		Append an opcode to this Header.
+
+		Returns the appended opcode.
 		"""
 		self._opcodes[opcode.name] = opcode
 		opcode.parent = self
+		return opcode
 
 	def append_subheader(self, subheader):
 		"""
-		Append a subheader to this Header
+		Append a subheader to this Header.
+
+		Returns the appended subheader.
 		"""
 		self._subheaders.append(subheader)
 		subheader.parent = self
+		return subheader
 
 	@property
 	def opcodes(self):
@@ -184,24 +209,36 @@ class Header(_SFZElement):
 	def iopcode(self, name):
 		"""
 		Returns an Opcode with the given name, if one exists in this Header or any of
-		its ancestors. Returns None if no such opcode exists.
+		its ancestors.
+
+		Returns None if no such opcode exists.
 		"""
 		return self._opcodes[name] if name in self._opcodes \
 			else None if self._parent is None \
 			else self._parent.iopcode(name)
 
 	def __getattr__(self, name):
-		try:
+		"""
+		Returns the value of the opcode with the given "name" contained in this Header,
+		or any of its ancestors.
+
+		Raises AttributeError
+		"""
+		if name[0] == '_':
 			return super().__getattribute__(name)
-		except AttributeError as err:
-			opcode = self.iopcode(name)
-			if opcode:
-				return opcode.value
-			if normal_opcode(name):
-				return None
-			raise err
+		opcode = self.iopcode(name)
+		if opcode:
+			return opcode.value
+		if normal_opcode(name):
+			return None
+		raise AttributeError(name)
 
 	def __setattr__(self, name, value):
+		"""
+		Sets the value of the opcode with the given "name", if it exists.
+
+		Creates the opcode inside this Header, if it does not yet exist.
+		"""
 		if name[0] == '_':
 			super().__setattr__(name, value)
 		elif name in self.__dict__:
@@ -339,9 +376,12 @@ class Region(Header):
 	"""
 
 	def may_contain(self, header):
+		"""
+		Used during parsing to determine where to append a newly parsed Header
+		"""
 		return type(header) not in [Global, Master, Group, Region]
 
-	def is_triggerd_by(self, key=None, lokey=None, hikey=None, lovel=None, hivel=None):
+	def is_triggerd_by(self, key = None, lokey = None, hikey = None, lovel = None, hivel = None):
 		"""
 		Returns boolean True/False if this Region matches the given criteria.
 		For example, to test if this region plays Middle C at any velocity:
@@ -362,6 +402,14 @@ class Region(Header):
 			return False
 		return True
 
+	@property
+	def sample(self):
+		"""
+		Returns the Sample opcode contained in this <region>
+		Returns None if no Sample has yet defined.
+		"""
+		return self._
+
 
 class Control(Header):
 	"""
@@ -415,6 +463,18 @@ class Opcode(_SFZElement):
 		self.name = name
 		self.value = value
 
+	def __deepcopy__(self, memo):
+		cls = self.__class__
+		result = cls.__new__(cls, self.name, self.value)
+		result.name = self.name
+		if cls is Sample:
+			result.basedir = self.basedir
+			result._value = self.abspath
+		else:
+			result._value = self._value
+		memo[id(self)] = result
+		return result
+
 	@property
 	def value(self):
 		"""
@@ -425,7 +485,8 @@ class Opcode(_SFZElement):
 	@value.setter
 	def value(self, value):
 		"""
-		Converts the given value to the type defined in the opcode definition.
+		Set the value of this Opcode, after converting the given "value" to
+		the data type of this Opcode, (see "type()").
 		"""
 		if self.type is float:
 			self._value = float(value)
@@ -442,6 +503,9 @@ class Opcode(_SFZElement):
 
 	@cached_property
 	def type(self):
+		"""
+		Returns the data type (i.e. float, int, or str) of this Opcode.
+		"""
 		return data_type(self.name)
 
 	@cached_property
@@ -493,7 +557,7 @@ class Opcode(_SFZElement):
 		return '%s=%s' % (self.name, self._value)
 
 	def __repr__(self):
-		return f'Opcode {self}'
+		return f'{self.__class__} {self}'
 
 	def write(self, stream):
 		"""
@@ -512,9 +576,16 @@ class Sample(Opcode):
 
 	def __init__(self, name, value, meta = None, basedir = None):
 		"""
-		When instantiating a Sample, the "name" and "_value" of the given path
-		is set in Opcode.__init__(). Afterwards, the "path" of the Sample may be
-		manipulated without destroying the initial value.
+		When instantiating a Sample, the "name" is always "Sample"; the "value"
+		is the path to the sample as it appears in the SFZ. In Opcode.__init__(),
+		the internal "_value" property is set to the given value.
+
+		The "meta" argument is passed to the constructor during parsing of an existing
+		SFZ file. You can safely ignore it when constructing an SFZ from scratch.
+
+		"basedir" would be the name of the directory containing the SFZ which owns this
+		Sample. This value is used for determing the relative path to a sample defined
+		with an absolute path.
 		"""
 		super().__init__(name, value, meta)
 		self.basedir = basedir
@@ -554,13 +625,24 @@ class Sample(Opcode):
 		"""
 		self._value = self.abspath
 
+	def _new_target(self, sfz_directory, samples_path):
+		"""
+		Returns a tuple ((str) value, (str) abspath),
+		where "value" is the new opcode value for this Sample, and
+		"abspath" is the target of a copy/move/symlink/hardlink op.
+		"""
+		return (
+			join(samples_path, self.basename),
+			join(sfz_directory, samples_path, self.basename)
+		)
+
 	def resolve_from(self, sfz_directory):
 		"""
 		Directs this Sample to use a relative path when writing .sfz.
 
 		"sfz_directory" is the directory in which the .sfz file is to be written.
 		"""
-		self.path = relpath(self.abspath, sfz_directory)
+		self._value = relpath(realpath(self.abspath), realpath(sfz_directory))
 
 	def copy_to(self, sfz_directory, samples_path):
 		"""
@@ -569,10 +651,14 @@ class Sample(Opcode):
 
 		"sfz_directory" is the directory in which the .sfz file is to be written.
 
-		"samples_path" must be a path relative the directory in which the .sfz
-		file is to be written.
+		"samples_path" is the directory where the samples are to be written, relative
+		to "sfz_directory".
 		"""
-		copy(self.abspath, self._fix_to_samples_dir(sfz_directory, samples_path))
+		value, path = self._new_target(sfz_directory, samples_path)
+		if exists(path):
+			unlink(path)
+		copy(self.abspath, path)
+		self._value = value
 
 	def move_to(self, sfz_directory, samples_path):
 		"""
@@ -581,10 +667,14 @@ class Sample(Opcode):
 
 		"sfz_directory" is the directory in which the .sfz file is to be written.
 
-		"samples_path" must be a path relative the directory in which the .sfz
-		file is to be written.
+		"samples_path" is the directory where the samples are to be written, relative
+		to "sfz_directory".
 		"""
-		move(self.abspath, self._fix_to_samples_dir(sfz_directory, samples_path))
+		value, path = self._new_target(sfz_directory, samples_path)
+		if exists(path):
+			unlink(path)
+		move(self.abspath, path)
+		self._value = value
 
 	def symlink_to(self, sfz_directory, samples_path):
 		"""
@@ -593,10 +683,14 @@ class Sample(Opcode):
 
 		"sfz_directory" is the directory in which the .sfz file is to be written.
 
-		"samples_path" must be a path relative the directory in which the .sfz
-		file is to be written.
+		"samples_path" is the directory where the samples are to be written, relative
+		to "sfz_directory".
 		"""
-		symlink(self.abspath, self._fix_to_samples_dir(sfz_directory, samples_path))
+		value, path = self._new_target(sfz_directory, samples_path)
+		if exists(path):
+			unlink(path)
+		symlink(realpath(self.abspath), realpath(path))
+		self._value = value
 
 	def hardlink_to(self, sfz_directory, samples_path):
 		"""
@@ -605,24 +699,14 @@ class Sample(Opcode):
 
 		"sfz_directory" is the directory in which the .sfz file is to be written.
 
-		"samples_path" must be a path relative the directory in which the .sfz
-		file is to be written.
+		"samples_path" is the directory where the samples are to be written, relative
+		to "sfz_directory".
 		"""
-		link(self.abspath, self._fix_to_samples_dir(sfz_directory, samples_path))
-
-	def _fix_to_samples_dir(self, sfz_directory, samples_path):
-		"""
-		Sets the "value" of this opcode to "<samples_path>/<sample basename>".
-
-		Returns the absolute path of the sample in the new samples_path.
-
-		"sfz_directory" is the directory in which the .sfz file is to be written.
-
-		"samples_path" must be a path relative the directory in which the .sfz
-		file is to be written.
-		"""
-		self._value = join(samples_path, self.basename)
-		return join(sfz_directory, samples_path, self.basename)
+		value, path = self._new_target(sfz_directory, samples_path)
+		if exists(path):
+			unlink(path)
+		hardlink(self.abspath, path)
+		self._value = value
 
 
 class Define(_Modifier):
